"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.BinaryPropertyTableModel = void 0;
const BinaryPropertyModels_1 = require("./BinaryPropertyModels");
const BinaryMetadataEntityModel_1 = require("./BinaryMetadataEntityModel");
const MetadataEntityModels_1 = require("../MetadataEntityModels");
const MetadataError_1 = require("../MetadataError");
/**
 * Implementation of the `PropertyTableModel` interface that is backed
 * by binary data.
 *
 * @internal
 */
class BinaryPropertyTableModel {
    constructor(binaryPropertyTable) {
        /**
         * A mapping from property IDs to the `PropertyModel`
         * instances that provide the property values. These
         * are the "columns" of the table
         */
        this._propertyIdToModel = {};
        this._binaryPropertyTable = binaryPropertyTable;
        // Initialize the `PropertyModel` instances for
        // the property table properties
        const propertyTable = this._binaryPropertyTable.propertyTable;
        const propertyTableProperties = propertyTable.properties;
        if (propertyTableProperties) {
            for (const propertyId of Object.keys(propertyTableProperties)) {
                const propertyModel = BinaryPropertyModels_1.BinaryPropertyModels.createPropertyModel(this._binaryPropertyTable, propertyId);
                this._propertyIdToModel[propertyId] = propertyModel;
            }
        }
        const metadataClass = binaryPropertyTable.metadataClass;
        this._semanticToPropertyId =
            MetadataEntityModels_1.MetadataEntityModels.computeSemanticToPropertyIdMapping(metadataClass);
    }
    /** {@inheritDoc PropertyTableModel.getMetadataEntityModel} */
    getMetadataEntityModel(index) {
        const propertyTable = this._binaryPropertyTable.propertyTable;
        const count = propertyTable.count;
        if (index < 0 || index >= count) {
            const message = `The index must be in [0,${count}), but is ${index}`;
            throw new MetadataError_1.MetadataError(message);
        }
        const semanticToPropertyId = this._semanticToPropertyId;
        const metadataEntityModel = new BinaryMetadataEntityModel_1.BinaryMetadataEntityModel(this, index, semanticToPropertyId);
        return metadataEntityModel;
    }
    /** {@inheritDoc PropertyTableModel.getPropertyModel} */
    getPropertyModel(propertyId) {
        return this._propertyIdToModel[propertyId];
    }
    /** {@inheritDoc PropertyTableModel.getClassProperty} */
    getClassProperty(propertyId) {
        const binaryPropertyTable = this._binaryPropertyTable;
        const metadataClass = binaryPropertyTable.metadataClass;
        const classProperties = metadataClass.properties;
        if (!classProperties) {
            return undefined;
        }
        return classProperties[propertyId];
    }
    /** {@inheritDoc PropertyTableModel.getPropertyTableProperty} */
    getPropertyTableProperty(propertyId) {
        const binaryPropertyTable = this._binaryPropertyTable;
        const propertyTable = binaryPropertyTable.propertyTable;
        const propertyTableProperties = propertyTable.properties;
        if (!propertyTableProperties) {
            return undefined;
        }
        return propertyTableProperties[propertyId];
    }
}
exports.BinaryPropertyTableModel = BinaryPropertyTableModel;
