"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TilesetStageExecutor = void 0;
const ContentStageExecutor_1 = require("./ContentStageExecutor");
const PipelineError_1 = require("./PipelineError");
const TilesetStages_1 = require("./TilesetStages");
const BasicTilesetProcessor_1 = require("../tilesetProcessing/BasicTilesetProcessor");
const TilesetUpgrader_1 = require("../tilesetProcessing/TilesetUpgrader");
const TilesetCombiner_1 = require("../tilesetProcessing/TilesetCombiner");
const ContentDataTypeChecks_1 = require("../contentTypes/ContentDataTypeChecks");
const ContentDataTypes_1 = require("../contentTypes/ContentDataTypes");
/**
 * Methods to execute `TilesetStage` objects.
 */
class TilesetStageExecutor {
    /**
     * Executes the given `TilesetStage`.
     *
     * @param tilesetStage - The `TilesetStage` object
     * @param currentInput - The current input name, or a temporary
     * name for intermediate steps (see `Pipeline.input` for details)
     * @param currentOutput - The current output name, or a temporary
     * name for intermediate steps (see `Pipeline.input` for details)
     * @param overwrite - Whether outputs should be overwritten if
     * they already exist
     * @returns A promise that resolves when the process is finished
     * @throws PipelineError If one of the processing steps causes
     * an error.
     */
    static async executeTilesetStage(tilesetStage, currentInput, currentOutput, overwrite) {
        console.log(`  Executing tilesetStage : ${tilesetStage.name}`);
        console.log(`    currentInput:  ${currentInput}`);
        console.log(`    currentOutput: ${currentOutput}`);
        try {
            await TilesetStageExecutor.executeTilesetStageInternal(tilesetStage, currentInput, currentOutput, overwrite);
        }
        catch (e) {
            throw new PipelineError_1.PipelineError(`${e}`);
        }
    }
    /**
     * Implementation for `executeTilesetStage`.
     *
     * For details about the arguments, see `executeTilesetStage`.
     *
     * @param tilesetStage - The `TilesetStage` object
     * @param currentInput - The current input name
     * @param currentOutput - The current output name
     * @param overwrite - Whether outputs should be overwritten
     * @returns A promise that resolves when the process is finished
     * @throws Error If one of the processing steps causes
     * an error.
     */
    static async executeTilesetStageInternal(tilesetStage, currentInput, currentOutput, overwrite) {
        if (tilesetStage.name === TilesetStages_1.TilesetStages.TILESET_STAGE_UPGRADE) {
            const quiet = false;
            const tilesetUpgrader = new TilesetUpgrader_1.TilesetUpgrader(quiet);
            await tilesetUpgrader.upgrade(currentInput, currentOutput, overwrite);
        }
        else if (tilesetStage.name === TilesetStages_1.TilesetStages.TILESET_STAGE_COMBINE) {
            const externalTilesetDetector = ContentDataTypeChecks_1.ContentDataTypeChecks.createIncludedCheck(ContentDataTypes_1.ContentDataTypes.CONTENT_TYPE_TILESET);
            const tilesetCombiner = new TilesetCombiner_1.TilesetCombiner(externalTilesetDetector);
            await tilesetCombiner.combine(currentInput, currentOutput, overwrite);
        }
        else {
            await TilesetStageExecutor.executeTilesetContentStages(tilesetStage, currentInput, currentOutput, overwrite);
        }
    }
    /**
     * Execute all `ContentStage` objects in the given `TilesetStage`.
     *
     * For details about the arguments, see `executeTilesetStage`.
     *
     * @param tilesetStage - The `TilesetStage` object
     * @param currentInput - The current input name
     * @param currentOutput - The current output name
     * @param overwrite - Whether outputs should be overwritten
     * @returns A promise that resolves when the process is finished
     * @throws Error If one of the processing steps causes
     * an error.
     */
    static async executeTilesetContentStages(tilesetStage, currentInput, currentOutput, overwrite) {
        try {
            const tilesetProcessor = new BasicTilesetProcessor_1.BasicTilesetProcessor();
            await tilesetProcessor.begin(currentInput, currentOutput, overwrite);
            const contentStages = tilesetStage.contentStages;
            if (contentStages) {
                for (let c = 0; c < contentStages.length; c++) {
                    const contentStage = contentStages[c];
                    const message = `    Executing contentStage ${c} of ` +
                        `${contentStages.length}: ${contentStage.name}`;
                    console.log(message);
                    await ContentStageExecutor_1.ContentStageExecutor.executeContentStage(contentStage, tilesetProcessor);
                }
            }
            await tilesetProcessor.end();
        }
        catch (e) {
            throw new PipelineError_1.PipelineError(`${e}`);
        }
    }
}
exports.TilesetStageExecutor = TilesetStageExecutor;
