"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TilesetStages = void 0;
const defined_1 = require("../base/defined");
const DeveloperError_1 = require("../base/DeveloperError");
const ContentStages_1 = require("./ContentStages");
/**
 * Methods to create `TilesetStage` objects.
 */
class TilesetStages {
    /**
     * Creates a tileset stage that performs the "upgrade" operation
     *
     * @returns The tileset stage
     */
    static createUpgrade() {
        const tilesetStage = {
            name: TilesetStages.TILESET_STAGE_UPGRADE,
            description: "Upgrade the input tileset to the latest version",
        };
        return tilesetStage;
    }
    /**
     * Creates a tileset stage that performs the "combine" operation
     *
     * @returns The tileset stage
     */
    static createCombine() {
        const tilesetStage = {
            name: TilesetStages.TILESET_STAGE_COMBINE,
            description: "Combine all external tilesets into one",
        };
        return tilesetStage;
    }
    /**
     * Creates a tileset stage from the given parameters.
     *
     * @param name - The `name` of the tileset stage
     * @param description - The `description` of the tileset stage
     * @param contentStages - The content stages
     * @returns The tileset stage
     */
    static create(name, description, contentStages) {
        const tilesetStage = {
            name: name,
            description: description,
            contentStages: contentStages,
        };
        return tilesetStage;
    }
    /**
     * Creates a `TilesetStage` object from the given (untyped) JSON.
     *
     * @param tilesetStageJson - The JSON object
     * @returns The `TilesetStage` object
     * @throws DeveloperError When the input was not valid
     */
    static createTilesetStage(tilesetStageJson) {
        if (typeof tilesetStageJson === "string") {
            const tilesetStage = {
                name: tilesetStageJson,
                contentStages: [],
            };
            return tilesetStage;
        }
        if (!(0, defined_1.defined)(tilesetStageJson.name)) {
            throw new DeveloperError_1.DeveloperError("The tilesetStage JSON does not define a name");
        }
        let contentStages = undefined;
        if (tilesetStageJson.contentStages) {
            contentStages = [];
            for (const contentStageJson of tilesetStageJson.contentStages) {
                const contentStage = ContentStages_1.ContentStages.createContentStage(contentStageJson);
                contentStages.push(contentStage);
            }
        }
        const tilesetStage = {
            name: tilesetStageJson.name,
            contentStages: contentStages,
        };
        return tilesetStage;
    }
}
exports.TilesetStages = TilesetStages;
/**
 * The `name` that identifies the "upgrade" tileset stage
 */
TilesetStages.TILESET_STAGE_UPGRADE = "upgrade";
/**
 * The `name` that identifies the "combine" tileset stage
 */
TilesetStages.TILESET_STAGE_COMBINE = "combine";
