/// <reference types="node" />
import { CompositeTileData } from "./CompositeTileData";
import { TileData } from "./TileData";
import { TileDataLayout } from "./TileDataLayouts";
/**
 * Methods for handling 3D Tiles tile data.
 *
 * Methods for reading, creating, and writing data in
 * B3DM, I3DM, PNTS, and CMPT format.
 *
 */
export declare class TileFormats {
    /**
     * Returns whether the given buffer contains composite (CMPT) tile data.
     *
     * @param buffer - The buffer
     * @returns Whether the given buffer contains composite tile data
     */
    static isComposite(buffer: Buffer): boolean;
    /**
     * Reads a `CompositeTileData` object from the given buffer.
     *
     * The given buffer is assumed to contain valid tile data in CMPT
     * format. Whether the magic header of a buffer indicates CMPT
     * format can be checked with `isComposite`.
     *
     * @param buffer - The buffer
     * @returns The `CompositeTileData`
     * @throws TileFormatError If the given buffer did not contain
     * valid CMPT data.
     */
    static readCompositeTileData(buffer: Buffer): CompositeTileData;
    /**
     * Reads a `TileData` object from the given buffer.
     *
     * This method can be applied to a buffer that contains valid B3DM,
     * I3DM or PNTS data.
     *
     * @param buffer - The buffer
     * @returns The `TileData`
     * @throws TileFormatError If the given buffer did not contain
     * valid tile data.
     */
    static readTileData(buffer: Buffer): TileData;
    /**
     * Extracts a `TileData` object from the given buffer, based
     * on the given layout information.
     *
     * This method can be applied to a buffer that contains valid B3DM,
     * I3DM or PNTS data, with a tile data layout that was created
     * with the `TileDataLayouts.create` method.
     *
     * @param buffer - The buffer
     * @param tileDataLayout - The tile data layout
     * @return The `TileData`
     * @internal
     */
    static extractTileData(buffer: Buffer, tileDataLayout: TileDataLayout): {
        header: {
            magic: string;
            version: number;
            gltfFormat: number | undefined;
        };
        featureTable: {
            json: any;
            binary: Buffer;
        };
        batchTable: {
            json: any;
            binary: Buffer;
        };
        payload: Buffer;
    };
    /**
     * Convenience method to collect all GLB (binary glTF) buffers from
     * the given tile data.
     *
     * This can be applied to B3DM and I3DM tile data, as well as CMPT
     * tile data. (For PNTS, it will return an empty array). When the
     * given tile data is a composite (CMPT) tile data, and recursively
     * collect the buffer from its inner tiles.
     *
     * @param tileDataBuffer - The tile data buffer
     * @returns The array of GLB buffers
     */
    static extractGlbBuffers(tileDataBuffer: Buffer): Buffer[];
    /**
     * Implementation for `extractGlbBuffers`, called recursively.
     *
     * @param tileDataBuffer - The tile data buffer
     * @param glbBuffers The array of GLB buffers
     */
    private static extractGlbBuffersInternal;
    /**
     * Creates a Batched 3D Model (B3DM) `TileData` instance from
     * a buffer that is assumed to contain valid binary glTF (GLB)
     * data.
     *
     * @param glbData - The GLB data
     * @returns The `TileData`
     */
    static createDefaultB3dmTileDataFromGlb(glbData: Buffer): TileData;
    /**
     * Creates a Batched 3D Model (B3DM) `TileData` instance from
     * a buffer that is assumed to contain valid binary glTF (GLB)
     * data.
     *
     * @param glbData - The GLB data
     * @param featureTableJson - The feature table JSON
     * @param featureTableBinary - The feature table binary
     * @param batchTableJson - The batch table JSON
     * @param batchTableBinary - The batch table binary
     * @returns The `TileData`
     */
    static createB3dmTileDataFromGlb(glbData: Buffer, featureTableJson: any, featureTableBinary: Buffer | undefined, batchTableJson: any, batchTableBinary: Buffer | undefined): TileData;
    /**
     * Creates an Instanced 3D Model (I3DM) `TileData` instance from
     * a buffer that is assumed to contain valid binary glTF (GLB)
     * data. The resulting tile data will represent a single instance,
     * namely exactly the given GLB object.
     *
     * @param glbData - The GLB data
     * @param featureTableJson - The feature table JSON
     * @param featureTableBinary - The feature table binary
     * @param batchTableJson - The batch table JSON
     * @param batchTableBinary - The batch table binary
     * @returns The `TileData`
     */
    static createDefaultI3dmTileDataFromGlb(glbData: Buffer): TileData;
    /**
     * Creates an Instanced 3D Model (I3DM) `TileData` instance from
     * a buffer that is assumed to contain valid binary glTF (GLB)
     * data.
     *
     * If no feature table information is given, then the resulting tile
     * data will represent a single instance, namely exactly the given
     * GLB object.
     *
     * @param glbData - The GLB data
     * @param featureTableJson - The feature table JSON
     * @param featureTableBinary - The feature table binary
     * @param batchTableJson - The batch table JSON
     * @param batchTableBinary - The batch table binary
     * @returns The `TileData`
     */
    static createI3dmTileDataFromGlb(glbData: Buffer, featureTableJson: any, featureTableBinary: Buffer | undefined, batchTableJson: any, batchTableBinary: Buffer | undefined): TileData;
    /**
     * Creates a Composite Tile Data (CMPT) instance from the given
     * `TileData` instances.
     *
     * @param tileDatas - The `TileData` instances
     * @returns The `CompositeTileData`
     */
    static createCompositeTileData(tileDatas: TileData[]): CompositeTileData;
    /**
     * Creates a buffer from the given `TileData`.
     *
     * The resulting buffer will contain a representation of the
     * tile data that can directly be written to a file. The
     * method will ensure that the padding requirements for the
     * 3D Tiles tile formats are met.
     *
     * @param tileData - The `TileData`
     * @returns The buffer
     */
    static createTileDataBuffer(tileData: TileData): Buffer;
    /**
     * Creates a buffer from the given `CompositeTileData`.
     *
     * The resulting buffer will contain a representation of the
     * tile data that can directly be written to a file. The
     * method will ensure that the padding requirements for the
     * 3D Tiles tile formats are met.
     *
     * @param compositeTileData - The `CompositeTileData`
     * @returns The buffer
     */
    static createCompositeTileDataBuffer(compositeTileData: CompositeTileData): Buffer;
}
//# sourceMappingURL=TileFormats.d.ts.map