import { ResourceResolver } from "../io/ResourceResolver";
import { Tileset } from "../structure/Tileset";
import { Schema } from "../structure/Metadata/Schema";
import { TraversedTile } from "./TraversedTile";
/**
 * Internal utility methods for tileset traversal, used for
 * the `TilesetTraverser` implementation.
 *
 * @internal
 */
export declare class TilesetTraversers {
    /**
     * Create the nodes that are the roots of external tilesets
     * that are referred to by the given traversed tile.
     *
     * If the given tile does not have any contents or none of
     * them refers to a tileset, then an empty array is returned.
     *
     * @param baseUri - The URI against which content URI are resolved
     * in order to obtain an absolute URI. This is only used for the case
     * of package (3TZ or 3DTILES) content, to create a `TilesetSource`
     * from the absolute URI.
     * @param traversedTile - The `TraversedTile`
     * @returns The external tileset roots
     * @throws DataError If one of the externa tilesets or
     * its associated files could not be resolved.
     */
    static createExternalTilesetRoots(baseUri: string, traversedTile: TraversedTile): Promise<TraversedTile[]>;
    /**
     * Fetch the information that is required for creating the root
     * nodes of external tilesets from the given URI.
     *
     * If the given URI does not refer to an external tileset,
     * then `undefined` is returned.
     *
     * Otherwise, it will return the parsed `Tileset` object,
     * and the `ResourceResolver` that can be used to resolve
     * resources from this tileset.
     *
     * @param baseUri - The URI against which the given URI is resolved
     * in order to obtain an absolute URI. This is only used for the case
     * of package (3TZ or 3DTILES) content, to create a `TilesetSource`
     * from the absolute URI.
     * @param uri - The URI
     * @param resourceResolver - The `ResourceResolver`
     * @returns The tileset
     * @throws DataError If an external tileset could not be
     * resolved or parsed.
     */
    private static resolveExternalTilesetContext;
    /**
     * Resolve the `Schema` for the given tileset.
     *
     * This is either the `tileset.schema`, or the schema that is
     * obtained from the `tileset.schemaUri`, or `undefined` if
     * neither of them are present.
     *
     * @param tileset - The tileset
     * @param resourceResolver - The `ResourceResolver` for loading
     * the schema from the `schemaUri` if necessary
     * @returns The `Schema`, or `undefined` if there is none
     * @throws DataError If the schema from the `schemaUri`
     * could not be resolved or parsed.
     */
    static resolveSchema(tileset: Tileset, resourceResolver: ResourceResolver): Promise<Schema | undefined>;
}
//# sourceMappingURL=TilesetTraversers.d.ts.map